/*
    SmartHomeGeeks.io Smart Battery Guard
    https://smarthomegeeks.io/battery-guard/
*/

#include <FS.h>                                 // FileShare used to store the credentials from WiFiManager
#include <ArduinoJson.h>                        // To create the JSON strings for MQTT messaging
#include <PubSubClient.h>                       // MQTT
#include <ArduinoOTA.h>                         // Over-the-Air firmware updates
#include <arduino_secrets.h>

// Device Configuration
const String deviceName = "Battery Guard";      // device name
const int analogInPin = A0;

// WiFi configuration
const char* ssid = SECRET_SSID;
const char* password = SECRET_PASS;

// MQTT configuration
const String mqtt_client = "battery_guard";            // MQTT client name
const char* mqtt_server = SECRET_MQTT_BROKER;          // MQTT broker IP
const char* mqtt_user = SECRET_MQTT_USERNAME;          // MQTT username
const char* mqtt_pass = SECRET_MQTT_PASSWORD;          // MQTT password
const int mqtt_port = 1883;                            // MQTT port (default: 1883)

bool keepAwake = true;                          // Deep sleep override
unsigned int interval = 60;                     // Deep sleep time in minutes
unsigned long lastMQTTconnect = 0;
unsigned long previousMillis = 0;
float conversion = 0.01715;

int totalRetainedMessages = 2; // Set the total number of retained messages expected
int processedRetainedMessages = 0; // Counter for processed retained messages

// WiFi Configuration
WiFiClient espClient;
PubSubClient client(espClient);

void setup_wifi() {
  // We start by connecting to a WiFi network
  Serial.println();
  Serial.print("Connecting to ");
  Serial.print(ssid);

  WiFi.mode(WIFI_STA);
  WiFi.setHostname(mqtt_client.c_str());
  WiFi.begin(ssid, password);

  int trycount = 0;
  while (WiFi.status() != WL_CONNECTED && trycount <= 10) {
    delay(500);
    Serial.print(".");
    trycount++;
  }
  Serial.println();

  // Still not connected? Go to sleep!
  if (WiFi.status() != WL_CONNECTED) {
    Serial.println("Unable to connect to WiFi");
    sleep();
  } else {
    Serial.println("WiFi connected");
    Serial.print("IP address: ");
    Serial.println(WiFi.localIP());
  }
}

float readBattery() {
  uint32_t value = 0;
  int rounds = 11;

  for (int i = 1; i <= rounds; i++) {
    value += analogRead(analogInPin);
  }
  value /= (uint32_t)rounds;
  return value * conversion;
}

void connectmqtt() {
  // prevent constant reconnect
  if (millis() - lastMQTTconnect < 5000) return;
  lastMQTTconnect = millis();

  if (client.connect(mqtt_client.c_str(), mqtt_user, mqtt_pass, (mqtt_client + "/availability").c_str(), 1, true, "offline")) {
    Serial.println("MQTT connected");
    client.setBufferSize(1024);
    
    // only publish autoDiscovery when not woken from sleep
    if (ESP.getResetReason() != String("Deep-Sleep Wake")) autoDiscovery();

    client.subscribe((mqtt_client + "/keepAwake").c_str());
    client.subscribe((mqtt_client + "/interval").c_str());
    client.subscribe((mqtt_client + "/calibrate").c_str());
    client.publish((mqtt_client + "/availability").c_str(), "online", true);
    client.publish((mqtt_client + "/ip").c_str(),  WiFi.localIP().toString().c_str(), true);

  } else {
    Serial.print("Failed, rc=");
    Serial.print(client.state());
    Serial.println(" try again in 5 seconds");
  }  
}

void saveConfig() {
  //save the custom parameters to FS
    DynamicJsonDocument json(1024);
    json["keepAwake"] = keepAwake;
    json["interval"] = interval;
    json["conversion"] = conversion;

    File configFile = SPIFFS.open("/config.json", "w");
    if (!configFile) {
      Serial.println("Failed to open config file for writing");
      return;
    }

    serializeJson(json, configFile);
    configFile.close();

    Serial.print("Config saved: ");
    serializeJson(json, Serial);
    Serial.println();
}

void callback(char* topic, byte* payload, unsigned int length) {
  Serial.print("\nTopic arrived: ");
  Serial.println(String(topic));
  
  Serial.print("Message arrived: ");
  for (int i=0;i<length;i++) {
    Serial.print((char)payload[i]);
  }
  Serial.println();

  // Increment the processedRetainedMessages counter
  processedRetainedMessages++;

  /// PROCESS THE INCOMING COMMANDS FOR keepAwake and interval
  if (String(topic) == mqtt_client + "/keepAwake") {
    keepAwake = (!strncmp((char*)payload, "ON", length) ? true : false);
    Serial.print("Changed keepAwake value to: ");
    Serial.println(String(keepAwake));
  }

  else if (String(topic) == mqtt_client + "/interval") {
    payload[length] = '\0';
    String s = String((char*)payload);
    interval = s.toInt();
    Serial.print("Changed sleep interval to: ");
    Serial.print(String(interval));
    Serial.println(" minutes");
  }

  else if (String(topic) == mqtt_client + "/calibrate") {
    payload[length] = '\0';
    String s = String((char*)payload);
    float calibrateV = s.toFloat();

    if (calibrateV == 0) {
      Serial.println("Empty calibrate value. Ignore");
      return;
    }

    Serial.print("Received new calibration value: ");
    Serial.println(calibrateV);

    Serial.print("Current conversion rate: ");
    Serial.println(conversion, 6);

    conversion = conversion / readBattery() * calibrateV;
    client.publish((mqtt_client + "/calibrate").c_str(), "0", true);

    Serial.print("New conversion rate: ");
    Serial.println(conversion, 6);

    // reset previousMillis to trigger a publish
    previousMillis = 0;
  }

  // save new configuration
  saveConfig();

}

void sleep () {
  Serial.println();

  // publish availability to offline
  client.publish((mqtt_client + "/availability").c_str(), "offline", true);

  // disconnect MQTT
  if (client.connected()) {
    Serial.println("Disconnecting MQTT...");
    client.disconnect();
  }

  // disconnect WiFi
  if (WiFi.status() == WL_CONNECTED) {
    Serial.println("Disconnecting WiFi...");
    WiFi.disconnect();
  }

  Serial.print("Going to sleep for: ");
  Serial.print(interval);
  Serial.println(" minutes");

  int sleepTime = (interval > 0 ? interval : 1);
  ESP.deepSleep(interval * 60000000);
}

void autoDiscovery () {
  Serial.println("Publishing Home Assistant Autodiscovery...");

  char buffer[1024];

  DynamicJsonDocument deviceData(512);
  deviceData["name"] = deviceName;
  deviceData["ids"] = mqtt_client;
  deviceData["mf"] = "SmartHomeGeeks.io";
  deviceData["mdl"] = "Battery Guard";
  deviceData["sw"] = "2023.09";
  
  // Switch for Deepsleep override
  DynamicJsonDocument root(1024);
  root["name"] = "Keep Awake";
  root["obj_id"] = mqtt_client + "_keepAwake";
  root["ic"] = "mdi:sleep-off";
  root["uniq_id"] = "switch." + mqtt_client;
  root["stat_t"] = mqtt_client + "/keepAwake";
  root["cmd_t"] = mqtt_client + "/keepAwake";
  root["ent_cat"] = "config";
  root["retain"] = true;
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/switch/" + mqtt_client + "/config").c_str(), buffer, true);
  root.clear();

  // Input for Calibrate
  root["name"] = "Calibrate";
  root["obj_id"] = mqtt_client + "_calibrate";
  root["ic"] = "mdi:gauge";
  root["uniq_id"] = "number." + mqtt_client + "_calibrate";
  root["stat_t"] = mqtt_client + "/calibrate";
  root["cmd_t"] = mqtt_client + "/calibrate";
  root["min"] = 0;
  root["max"] = 9999;
  root["step"] = 0.01;
  root["mode"] = "box";
  root["unit_of_meas"] = "V";
  root["avty_t"] = mqtt_client + "/availability";
  root["ent_cat"] = "config";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/number/" + mqtt_client + "/calibrate/config").c_str(), buffer, true);
  root.clear();

  // Input for Deepsleep time
  root["name"] = "Interval";
  root["obj_id"] = mqtt_client + "_interval";
  root["ic"] = "mdi:timer";
  root["uniq_id"] = "number." + mqtt_client + "_interval";
  root["stat_t"] = mqtt_client + "/interval";
  root["cmd_t"] = mqtt_client + "/interval";
  root["min"] = 1;
  root["max"] = 999;
  root["mode"] = "box";
  root["unit_of_meas"] = "min";
  root["ent_cat"] = "config";
  root["retain"] = true;
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/number/" + mqtt_client + "/interval/config").c_str(), buffer, true);
  root.clear();

  // sensors
  root["name"] = "Battery Voltage";
  root["obj_id"] = mqtt_client + "_battery_voltage";
  root["stat_t"] = mqtt_client + "/battery_voltage";
  root["unit_of_meas"] = "V";
  root["ic"] = "mdi:car-battery";
  root["uniq_id"] = "sensor." + mqtt_client + "_battery_voltage";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/sensor/" + mqtt_client + "/battery_voltage/config").c_str(), buffer, true);
  root.clear();

  // RSSI sensor
  root["name"] = "RSSI";
  root["obj_id"] = mqtt_client + "_rssi";
  root["stat_t"] = mqtt_client + "/rssi";
  root["unit_of_meas"] = "dBm";
  root["ic"] = "mdi:access-point";
  root["uniq_id"] = "sensor." + mqtt_client + "_rssi";
  root["avty_t"] = mqtt_client + "/availability";
  root["ent_cat"] = "diagnostic";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/sensor/" + mqtt_client + "/rssi/config").c_str(), buffer, true);
  root.clear();

  // IP
  root["name"] = "IP";
  root["obj_id"] = mqtt_client + "_ip";
  root["stat_t"] = mqtt_client + "/ip";
  root["ic"] = "mdi:ip-network";
  root["uniq_id"] = "sensor." + mqtt_client + "_ip";
  root["avty_t"] = mqtt_client + "/availability";
  root["ent_cat"] = "diagnostic";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/sensor/" + mqtt_client + "/ip/config").c_str(), buffer, true);
  root.clear();

  // publish saved settings
  client.publish((mqtt_client + "/keepAwake").c_str(), (keepAwake ? "ON" : "OFF"), true);
  client.publish((mqtt_client + "/interval").c_str(), String(interval).c_str(), true);
  client.publish((mqtt_client + "/calibrate").c_str(), "0", true);

}

void setup() {
  // put your setup code here, to run once:
  Serial.begin(115200);
  Serial.println();

  //read configuration from FS json
  Serial.println("Mounting FS...");

  if (SPIFFS.begin()) {
    Serial.println("Mounted file system");
    if (SPIFFS.exists("/config.json")) {
      //file exists, reading and loading
      Serial.println("Reading config file");
      File configFile = SPIFFS.open("/config.json", "r");
      if (configFile) {
        Serial.println("Opened config file");
        size_t size = configFile.size();
        // Allocate a buffer to store contents of the file.
        std::unique_ptr<char[]> buf(new char[size]);

        configFile.readBytes(buf.get(), size);

        DynamicJsonDocument json(1024);
        auto deserializeError = deserializeJson(json, buf.get());
        if ( ! deserializeError ) {
          keepAwake = (json["keepAwake"] != nullptr ? json["keepAwake"] : keepAwake);
          interval = (json["interval"] != nullptr ? json["interval"] : interval);
          conversion = (json["conversion"] != nullptr ? json["conversion"] : conversion) ;
        } else {
          Serial.println("Failed to load json config");
        }
        configFile.close();
      }
    }
  } else {
    Serial.println("Failed to mount FS");
  }
  //end read

  // Configure OTA
  ArduinoOTA.setHostname(mqtt_client.c_str());
  ArduinoOTA.begin(); 

  // Setup WiFi
  setup_wifi();

  // Connect MQTT
  client.setServer(mqtt_server, mqtt_port);
  client.setCallback(callback);
}

void loop() {
  // put your main code here, to run repeatedly:
  if (!client.connected()) connectmqtt();

  unsigned long currentMillis = millis();
  if (client.connected() && (previousMillis == 0 || currentMillis < previousMillis || currentMillis - previousMillis >= 60000)) {
    previousMillis = currentMillis;

    float sensorValue = readBattery();
    Serial.print("Publishing battery voltage: ");
    Serial.println(sensorValue);

    client.publish((mqtt_client + "/battery_voltage").c_str(), String(sensorValue).c_str(), true);
    client.publish((mqtt_client + "/rssi").c_str(),  String(WiFi.RSSI()).c_str(), true);
  }

  if (!keepAwake && processedRetainedMessages >= totalRetainedMessages) sleep();

  ArduinoOTA.handle();

  client.loop();

}
